/*	analyze.c 
 * ------------------------------------------
 *	This program analyzes data recorded by the included MPI
 *	pingpong utility.  It creates a point for each task recorded by 
 *	pingpong, and randomly places it in an x,y system.  then, via
 *	simulated annealing, the nodes are shaped into an approximate
 *	best-fit graph, and their positions are recorded
 *
 *	by Sean Finney <finney@cs.swarthmore.edu>
 */
// ----------------------------------------------------------------- 
//	#includes
// ----------------------------------------------------------------- 
#include <stdio.h>
#include <stdlib.h>		// rand
#include <time.h>		// time
#include <sys/types.h>	// time
#include <math.h>	// pow


// ----------------------------------------------------------------- 
//	#defines
// ----------------------------------------------------------------- 

#define NUM_ITERATIONS	10000.0

#define USER_ERR 		-1		// these are for bail
#define INTERNAL_ERR	-2

#define DIST(x1, y1, x2, y2)	(hypot(x1-x2, y1-y2))
#define PROCDIST(p1, p2)	(DIST(p1.x, p1.y, p2.x, p2.y))
#define PROCTMPDIST(p1, p2)	(DIST(p1.tmp_x, p1.tmp_y, p2.tmp_x, p2.tmp_y))

#define INITIAL_T NUM_ITERATIONS
#define T_ALPHA 3.25
#define T_REDUCE(k, a) (INITIAL_T*pow((1.0-(k/INITIAL_T)), a))
#define T_REDUCE_INTERVAL (100)


struct proc {
	short rank;			// just an identifier
	double x;			// x position
	double y;			// y position
	double tmp_x;		// for holding new coordinate until calcs
	double tmp_y;		//	are completed
}; 
typedef struct proc proc_t;



// ----------------------------------------------------------------- 
// extern/global variables
// ----------------------------------------------------------------- 

char *g_progname;
char *g_filename;
int g_numprocs;
double **g_latencies;
double **g_distances;
proc_t *g_procs;

// these are the cmdline configgable settings
short g_record_enstates=0;
short g_record_beforefile=0;
short g_bequiet=0;
char *g_afterfile_name;

// these are just for the time being while I figure out the best method of
//	going about this
int g_iter_num=0;
double g_lowest_estate=0;
double g_current_estate=0;
int g_iter_val_at_lowest=-1;

// ----------------------------------------------------------------- 
//	function prototypes
// ----------------------------------------------------------------- 

void load_datafile(void);
void init_procs(void);
void calculate_distances(void);
void single_step(int temp);
int eval_position(int procnum, int num_procs);
void print_coords(FILE *f);
void print_distances(void);
void parse_cmdline(int count, char* args[]);
void bail(char *msg, int status);
void usage(void);

// ----------------------------------------------------------------- 

int main(int argc, char *argv[]){
	FILE *beforefile, *afterfile;
	int i, numbigloops;
	double temp, alpha=T_ALPHA;

	srand48(time(NULL));
	parse_cmdline(argc, argv);
	if(g_record_beforefile)beforefile=fopen("before", "w");
	afterfile=fopen(g_afterfile_name, "w");
	
	load_datafile();
	init_procs();

	if(g_record_beforefile)print_coords(beforefile);

	for(numbigloops=0;numbigloops<24;numbigloops++){
		temp=INITIAL_T;
		for(i=0;i<NUM_ITERATIONS;i++){
			calculate_distances();
			single_step(i);
			if(i%T_REDUCE_INTERVAL==0) temp=T_REDUCE(i, alpha);
		}
	if(!g_bequiet){
		printf("\titeration: %-10.8d current e: %-8.6f lowest e: %-8.6f\r", 
			g_iter_num, g_current_estate, g_lowest_estate);
		fflush(stdout);
		}
	}
	if(!g_bequiet)printf("\n");
	print_coords(afterfile);

	if(g_record_beforefile) fclose(beforefile);

	if(!g_bequiet){
		if(g_record_beforefile)
			printf("state before annealing can be found in \"before\"\n");
		printf("state after annealing can be found in %s\n", g_afterfile_name);	
	}
	fclose(afterfile);
	return(0);
}

// ----------------------------------------------------------------- 

void load_datafile(void){
	int i, from, to, pwr;
	char *slat;
	double lat;
	FILE *infile;

	// XXX
	//	oops, i wasn't doing fscanf correctly.  anyhoo, this is a kludge,
	//	and if we ever have latencies more than 100 chars long, this
	//	will have to change :p
	slat=(char*)malloc(sizeof(char)*100);
	infile=fopen(g_filename, "r");
	if(!infile) bail("unable to open datafile!", INTERNAL_ERR);
	fscanf(infile, "Number Processors: %d\n", &g_numprocs);

	// malloc the 2D latency array
	g_latencies = (double**)malloc(sizeof(double*)*g_numprocs);
	for(i=0;i<g_numprocs;i++)
		g_latencies[i] = (double*)malloc(sizeof(double)*g_numprocs);

	// load all the latencies into the newly allocated 2D array
	while(fscanf(infile, "%d %d %[^\t]\t", &from, &to, slat)!=EOF) 
		g_latencies[from][to]=atof(slat);
	
	fclose(infile);
}

void init_procs(void){
	int i;
	
	g_procs = (proc_t*)malloc(sizeof(proc_t)*g_numprocs);
	for(i=0;i<g_numprocs;i++){
		g_procs[i].rank = i;
		g_procs[i].tmp_x = g_procs[i].x = (drand48() * 20) - 10;
		g_procs[i].tmp_y = g_procs[i].y = (drand48() * 20) - 10;
	}
	g_distances = (double**)malloc(sizeof(double*)*g_numprocs);
	for(i=0;i<g_numprocs;i++)g_distances[i]=(double*)malloc(sizeof(double)*g_numprocs);

}

void calculate_distances(void){
	short i,j; 

	for(i=0;i<g_numprocs;i++){
		for(j=i;j<g_numprocs;j++){
			if(i==j) g_distances[i][j] = g_distances[j][i] = 0;
			else{
				g_distances[i][j] = PROCDIST(g_procs[i], g_procs[j]);
				g_distances[j][i] = g_distances[i][j];
			}
		}
	}
}

void single_step(int temp){
	int i;

	for(i=0;i<g_numprocs;i++){
		g_procs[i].tmp_x += pow(temp/INITIAL_T,2)*((2.*drand48())-1.)*10.;
		g_procs[i].tmp_y += pow(temp/INITIAL_T,2)*((2.*drand48())-1.)*10.;
		if(eval_position(i, g_numprocs) 
		|| rand()>((double)RAND_MAX*pow(temp/INITIAL_T,0.0625))){
			g_procs[i].x =	g_procs[i].tmp_x;
			g_procs[i].y =	g_procs[i].tmp_y;
		} else {
			g_procs[i].tmp_x =	g_procs[i].x;
			g_procs[i].tmp_y =	g_procs[i].y;
		}
	}
	g_current_estate/=(double)g_numprocs;
	if(g_iter_val_at_lowest == -1 || g_current_estate < g_lowest_estate){
		g_lowest_estate = g_current_estate;
		g_iter_val_at_lowest = g_iter_num;
	} 

	if(g_record_enstates)	
		fprintf(stderr, "%d\t%g\n", g_iter_num, g_current_estate);
	g_iter_num++;
}

void print_coords(FILE *f){
	short i,j;
	if(!f) return;
	for(i=0;i<g_numprocs;i++){
		fprintf(f, "%g\t%g\n", g_procs[i].x, g_procs[i].y); 
	}
}

void print_distances(void){
	short i,j;
	for(i=0;i<g_numprocs;i++){
		for(j=i;j<g_numprocs;j++){
			printf("%d<-->%d:\t%g\n", i, j, g_distances[i][j]); 
		}
	}
	printf("\n");
}

int eval_position(int procnum, int num_procs){
	int i, j;
	double en_state_before=0, en_state_after=0;

	for(i=0;i<num_procs;i++){
		en_state_before+=fabs((g_distances[procnum][i])-(g_latencies[procnum][i]));
		en_state_after+=fabs(PROCTMPDIST(g_procs[procnum], g_procs[i])-
							  (g_latencies[procnum][i]));
	}
	en_state_after/=(double)g_numprocs;
	en_state_before/=(double)g_numprocs;
	/*
	if(g_iter_val_at_lowest == -1 || en_state_after < g_lowest_estate){
		g_lowest_estate = en_state_after;
		g_iter_val_at_lowest = g_iter_num;
	}
	*/
	g_current_estate+=((en_state_after<en_state_before)?en_state_after:en_state_before);
	return (en_state_after<en_state_before);
}

//----------------------------------------------------------------- 
void parse_cmdline(int count, char *args[]){
	int pos=1;
	g_progname = args[0];

	if(count < 2) bail("not enough args, fool!", USER_ERR);
	while(pos < count){
		if(strcmp("-b", args[pos])==0){ g_record_beforefile=1; }
		else if(strcmp("-e", args[pos])==0){ g_record_enstates=1; }
		else if(strcmp("-o", args[pos])==0){
			if(!args[++pos]) bail("-f requires a filename!", USER_ERR);
			g_afterfile_name=args[pos];
		}
		else if(strcmp("-q", args[pos])==0){ g_bequiet=1; }
		// last option, if it doesn't match, we have a filename
		else{
			if(pos!=count-1) {
				fprintf(stderr, 
				"unknown option %s, assuming you want this as an input file\n",
				args[pos]);
			}if(g_filename){
				fprintf(stderr, 
				"overwriting previous value for input file (%s) with %s\n", 
				g_filename, args[pos]);
			}
			g_filename = args[pos];
		}
		// if(strcmp("option", args[pos])==0){....}
		pos++; 
	}
	if(!g_afterfile_name){
		g_afterfile_name=(char*)malloc(sizeof(char)*(strlen("after.out")+1));
		strcpy(g_afterfile_name, "after.out");
	}
}

void bail(char *msg, int status){
	fprintf(stderr, "%s\n", msg);
	if(status==USER_ERR) usage();
	exit(status);
}

void usage(void){
fprintf(stderr, "%s: usage:\n", g_progname);
fprintf(stderr, "%s [-beq] [-o outputfile] datafile\n", g_progname);
}
